# Copyright (c) 2010-2012 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
from asyncio.exceptions import TimeoutError as AsyncioTimeoutError
import errno
import fcntl
import importlib.metadata
import logging
from types import TracebackType
from typing import TYPE_CHECKING

import eventlet
import eventlet.asyncio
import eventlet.debug
import eventlet.greenthread
import eventlet.hubs
import eventlet.hubs.asyncio
import eventlet.patcher

if TYPE_CHECKING:
    from typing_extensions import Self

    # Needed until we bump our minimum to Python 3.11
    #
    # https://github.com/python/typeshed/issues/7855
    from _typeshed import SupportsWrite

    _StreamHandler = logging.StreamHandler[SupportsWrite[str]]
else:
    _StreamHandler = logging.StreamHandler

# We want the blocking APIs, because we set file descriptors to non-blocking.
os = eventlet.patcher.original("os")
# Used to communicate between real threads:
SimpleQueue = eventlet.patcher.original("queue").SimpleQueue
# Real OS-level threads:
threading = eventlet.patcher.original("threading")


class _BaseMutex:
    """Shared code for different mutex implementations."""

    def __init__(self) -> None:
        self.owner: int | None = None
        self.recursion_depth = 0

    def acquire(self, blocking: bool = True) -> bool:
        """Acquire the mutex.

        If called with blocking=False, returns True if the mutex was
        acquired and False if it wasn't. Otherwise, blocks until the mutex
        is acquired and returns True.
        This lock is recursive; the same greenthread may acquire it as many
        times as it wants to, though it must then release it that many times
        too.
        """
        current_greenthread_id = id(eventlet.greenthread.getcurrent())
        if self.owner == current_greenthread_id:
            self.recursion_depth += 1
            return True

        return self._acquire_eventlet(blocking, current_greenthread_id)

    def _acquire_eventlet(
        self, blocking: bool, current_greenthread_id: int
    ) -> bool:
        raise NotImplementedError()

    def release(self) -> None:
        """Release the mutex."""
        current_greenthread_id = id(eventlet.greenthread.getcurrent())
        if self.owner != current_greenthread_id:
            raise RuntimeError("cannot release un-acquired lock")

        if self.recursion_depth > 0:
            self.recursion_depth -= 1
            return

        self.owner = None
        self._release_eventlet()

    def _release_eventlet(self) -> None:
        raise NotImplementedError()

    def close(self) -> None:
        """Close the mutex.

        This releases its file descriptors.
        You can't use a mutex after it's been closed.
        """
        self.owner = None
        self.recursion_depth = 0


class _ReallyPipeMutex(_BaseMutex):
    """Mutex using a pipe.

    Works across both greenlets and real threads, even at the same time.

    Class code copied from Swift's swift/common/utils.py
    Related eventlet bug: https://github.com/eventlet/eventlet/issues/432
    """

    def __init__(self) -> None:
        super().__init__()

        self.rfd, self.wfd = os.pipe()

        # You can't create a pipe in non-blocking mode; you must set it
        # later.
        rflags = fcntl.fcntl(self.rfd, fcntl.F_GETFL)
        fcntl.fcntl(self.rfd, fcntl.F_SETFL, rflags | os.O_NONBLOCK)
        os.write(self.wfd, b'-')  # start unlocked

        # Usually, it's an error to have multiple greenthreads all waiting
        # to read the same file descriptor. It's often a sign of inadequate
        # concurrency control; for example, if you have two greenthreads
        # trying to use the same memcache connection, they'll end up writing
        # interleaved garbage to the socket or stealing part of each others'
        # responses.
        #
        # In this case, we have multiple greenthreads waiting on the same
        # file descriptor by design. This lets greenthreads in real thread A
        # wait with greenthreads in real thread B for the same mutex.
        # Therefore, we must turn off eventlet's multiple-reader detection.
        #
        # It would be better to turn off multiple-reader detection for only
        # our calls to trampoline(), but eventlet does not support that.
        eventlet.debug.hub_prevent_multiple_readers(False)

    def _acquire_eventlet(
        self, blocking: bool, current_greenthread_id: int
    ) -> bool:
        while True:
            try:
                # If there is a byte available, this will read it and remove
                # it from the pipe. If not, this will raise OSError with
                # errno=EAGAIN.
                os.read(self.rfd, 1)
                self.owner = current_greenthread_id
                return True
            except OSError as err:
                if err.errno != errno.EAGAIN:
                    raise

                if not blocking:
                    return False

                # Tell eventlet to suspend the current greenthread until
                # self.rfd becomes readable. This will happen when someone
                # else writes to self.wfd.
                eventlet.hubs.trampoline(self.rfd, read=True)

    def _release_eventlet(self) -> None:
        os.write(self.wfd, b'X')

    def close(self) -> None:
        """Close the mutex.

        This releases its file descriptors.
        You can't use a mutex after it's been closed.
        """
        if self.wfd is not None:
            os.close(self.wfd)
            self.wfd = None
        if self.rfd is not None:
            os.close(self.rfd)
            self.rfd = None
        super().close()

    def __del__(self) -> None:
        # We need this so we don't leak file descriptors. Otherwise, if you
        # call get_logger() and don't explicitly dispose of it by calling
        # logger.logger.handlers[0].lock.close() [1], the pipe file
        # descriptors are leaked.
        #
        # This only really comes up in tests. Service processes tend to call
        # get_logger() once and then hang on to it until they exit, but the
        # test suite calls get_logger() a lot.
        #
        # [1] and that's a completely ridiculous thing to expect callers to
        # do, so nobody does it and that's okay.
        self.close()

    def __enter__(self) -> 'Self':
        self.acquire()
        return self

    def __exit__(
        self,
        type_: type[BaseException] | None,
        value: BaseException | None,
        traceback: TracebackType | None,
    ) -> None:
        self.release()


class _AsyncioMutex(_BaseMutex):
    """Alternative implementation of mutex for eventlet asyncio hub.

    When using the eventlet asyncio hub, multiple file descriptors as readers
    aren't supported.  So instead, we use two levels of locking:

        1. A normal RLock, for locking across OS threads.

        2. An ``asyncio.Lock`` for locking across greenlets within a single OS
           thread (each OS thread running greenlets has its own asyncio loop)
    """

    def __init__(self) -> None:
        super().__init__()
        self._asyncio_lock = asyncio.Lock()
        self._threading_lock = threading.RLock()

    async def _asyncio_acquire(
        self, blocking: bool, current_greenthread_id: int
    ) -> bool:
        if blocking:
            timeout = None
        else:
            timeout = 0.000001
        try:
            await asyncio.wait_for(
                self._asyncio_lock.acquire(), timeout=timeout
            )
        except AsyncioTimeoutError:
            return False
        else:
            self.owner = current_greenthread_id
            return True

    def _acquire_eventlet(
        self, blocking: bool, current_greenthread_id: int
    ) -> bool:
        return eventlet.asyncio.spawn_for_awaitable(  # type: ignore
            self._asyncio_acquire(blocking, current_greenthread_id)
        ).wait()

    def acquire(self, blocking: bool = True) -> bool:
        """Acquire the mutex.

        If called with blocking=False, returns True if the mutex was
        acquired and False if it wasn't. Otherwise, blocks until the mutex
        is acquired and returns True.
        This lock is recursive; the same greenthread may acquire it as many
        times as it wants to, though it must then release it that many times
        too.
        """
        # First, acquire the RLock:
        rlock_acquired = self._threading_lock.acquire(blocking=False)
        if not rlock_acquired and not blocking:
            return False
        # Simulate blocking while allowing other greenlets to run:
        while not rlock_acquired:
            rlock_acquired = self._threading_lock.acquire(blocking=False)
            # The chance of hitting this path is usually quite low (it requires
            # multiple _OS_ threads having a conflict) and sleeping for too
            # short a time leads to very slow results in one of the tests,
            # perhaps due to lots of context switching overhead. So compromise
            # on a short but not too-short sleep.
            eventlet.sleep(0.0001)
        # Then, do the eventlet locking:
        return super().acquire(blocking=blocking)

    def _release_eventlet(self) -> None:
        self._asyncio_lock.release()

    def release(self) -> None:
        """Release the mutex."""
        # We release in reverse order from acquire(), first eventlet and then
        # the RLock:
        super().release()
        self._threading_lock.release()

    def close(self) -> None:
        """Close the mutex."""
        del self._asyncio_lock
        del self._threading_lock
        super().close()


_HUB = eventlet.hubs.get_hub()
if isinstance(_HUB, eventlet.hubs.asyncio.Hub):
    major, minor, patch = map(
        int, importlib.metadata.version("eventlet").split(".")[:3]
    )
    if (major, minor, patch) < (0, 38, 2):
        raise RuntimeError(
            "eventlet 0.38.2 or later is required when using asyncio hub"
        )
    PipeMutex: type[_AsyncioMutex | _ReallyPipeMutex] = _AsyncioMutex
else:
    PipeMutex = _ReallyPipeMutex


def pipe_createLock(self: _StreamHandler) -> None:
    """Replacement for logging.Handler.createLock method."""
    self.lock = PipeMutex()  # type: ignore



#!/usr/bin/env python3
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import argparse
import collections
from collections.abc import Callable, Iterator, Sequence
import functools
import io
import sys
import time
from typing import cast

from oslo_serialization import jsonutils
from oslo_utils import importutils

from oslo_log.formatters import JSONLogRecord
from oslo_log import log

termcolor = importutils.try_import('termcolor')


_USE_COLOR = False
DEFAULT_LEVEL_KEY = 'levelname'
DEFAULT_TRACEBACK_KEY = 'traceback'


def main() -> None:
    global _USE_COLOR
    args = parse_args()
    _USE_COLOR = args.color
    formatter = functools.partial(
        console_format,
        args.prefix,
        args.locator,
        loggers=args.loggers,
        levels=args.levels,
        level_key=args.levelkey,
        traceback_key=args.tbkey,
    )
    if args.lines:
        # Read backward until we find all of our newline characters
        # or reach the beginning of the file
        args.file.seek(0, 2)
        newlines = 0
        pos = args.file.tell()
        while newlines <= args.lines and pos > 0:
            pos = pos - 1
            args.file.seek(pos)
            if args.file.read(1) == '\n':
                newlines = newlines + 1
    try:
        for line in reformat_json(args.file, formatter, args.follow):
            print(line)
    except KeyboardInterrupt:
        sys.exit(0)


def parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "file",
        nargs='?',
        default=sys.stdin,
        type=argparse.FileType(),
        help="JSON log file to read from (if not provided"
        " standard input is used instead)",
    )
    parser.add_argument(
        "--prefix",
        default='%(asctime)s.%(msecs)03d %(process)s %(levelname)s %(name)s',
        help="Message prefixes",
    )
    parser.add_argument(
        "--locator",
        default='[%(funcname)s %(pathname)s:%(lineno)s]',
        help="Locator to append to DEBUG records",
    )
    parser.add_argument(
        "--levelkey",
        default=DEFAULT_LEVEL_KEY,
        help="Key in the JSON record where the level is held",
    )
    parser.add_argument(
        "--tbkey",
        default=DEFAULT_TRACEBACK_KEY,
        help="Key in the JSON record where the traceback/exception is held",
    )
    parser.add_argument(
        "-c",
        "--color",
        action='store_true',
        default=False,
        help="Color log levels (requires `termcolor`)",
    )
    parser.add_argument(
        "-f",
        "--follow",
        action='store_true',
        default=False,
        help="Continue parsing new data until KeyboardInterrupt",
    )
    parser.add_argument(
        "-n",
        "--lines",
        required=False,
        type=int,
        help="Last N number of records to view."
        " (May show less than N records when used"
        " in conjuction with --loggers or --levels)",
    )
    parser.add_argument(
        "--loggers",
        nargs='*',
        default=[],
        help="only return results matching given logger(s)",
    )
    parser.add_argument(
        "--levels",
        nargs='*',
        default=[],
        help="Only return lines matching given log level(s)",
    )
    args = parser.parse_args()
    if args.color and not termcolor:
        raise ImportError(
            "Coloring requested but `termcolor` is not importable"
        )
    return args


def colorise(key: str, text: str | None = None) -> str:
    if text is None:
        text = key
    if not _USE_COLOR:
        return text
    colors = {
        'exc': ('red', ['reverse', 'bold']),
        'FATAL': ('red', ['reverse', 'bold']),
        'ERROR': ('red', ['bold']),
        'WARNING': ('yellow', ['bold']),
        'WARN': ('yellow', ['bold']),
        'INFO': ('white', ['bold']),
    }
    color, attrs = colors.get(key, ('', []))
    if color:
        return cast(str, termcolor.colored(text, color=color, attrs=attrs))
    return text


def warn(prefix: str, msg: object) -> str:
    return "{}: {}".format(colorise('exc', prefix), msg)


def reformat_json(
    fh: io.StringIO,
    formatter: Callable[..., Iterator[str]],
    follow: bool = False,
) -> Iterator[str]:
    # using readline allows interactive stdin to respond to every line
    while True:
        line = fh.readline()
        if not line:
            if follow:
                time.sleep(0.1)
                continue
            else:
                break
        line = line.strip()
        if not line:
            continue
        try:
            record = jsonutils.loads(line)
        except ValueError:
            yield warn("Not JSON", line)
            continue
        yield from formatter(record)


def console_format(
    prefix: str,
    locator: str,
    record: JSONLogRecord,
    loggers: Sequence[str] = [],
    levels: Sequence[str] = [],
    level_key: str = DEFAULT_LEVEL_KEY,
    traceback_key: str = DEFAULT_TRACEBACK_KEY,
) -> Iterator[str]:
    # Provide an empty string to format-specifiers the record is
    # missing, instead of failing. Doesn't work for non-string
    # specifiers.
    record = collections.defaultdict(str, record)  # type: ignore
    # skip if the record doesn't match a logger we are looking at
    if loggers:
        name = record['name']
        if not any(name.startswith(n) for n in loggers):
            return

    if levels:
        if record.get(level_key) not in levels:
            return

    levelname = record.get(level_key)
    if levelname:
        record[level_key] = colorise(levelname)  # type: ignore

    try:
        prefix = prefix % record
    except TypeError:
        # Thrown when a non-string format-specifier can't be filled in.
        # Dict comprehension cleans up the output
        yield warn(
            'Missing non-string placeholder in record',
            {
                str(k): str(v) if isinstance(v, str) else v
                for k, v in record.items()
            },
        )
        return

    locator = ''
    if record.get('levelno', 100) <= log.DEBUG or levelname == 'DEBUG':
        locator = locator % record

    yield ' '.join(x for x in [prefix, record['message'], locator] if x)

    tb = record.get(traceback_key)
    if tb:
        if type(tb) is str:
            tb = tb.rstrip().split("\n")
        for tb_line in tb:  # type: ignore
            yield ' '.join([prefix, tb_line])


if __name__ == '__main__':
    main()



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import annotations

import inspect
import logging
import logging.config
import logging.handlers
import os
from typing import TYPE_CHECKING

try:
    from systemd import journal  # type: ignore [import-not-found]
except ImportError:
    journal = None
try:
    import syslog
except ImportError:
    syslog = None  # type: ignore

if TYPE_CHECKING:
    # Needed until we bump our minimum to Python 3.11
    #
    # https://github.com/python/typeshed/issues/7855
    from _typeshed import SupportsWrite

    _StreamHandler = logging.StreamHandler[SupportsWrite[str]]
else:
    _StreamHandler = logging.StreamHandler

NullHandler = logging.NullHandler

_AUDIT = logging.INFO + 1
_TRACE = 5

# This is a copy of the numerical constants from syslog.h. The
# definition of these goes back at least 20 years, and is specifically
# 3 bits in a packed field, so these aren't likely to ever need
# changing.
SYSLOG_MAP = {
    "CRITICAL": 2,
    "ERROR": 3,
    "WARNING": 4,
    "WARN": 4,
    "INFO": 6,
    "DEBUG": 7,
}


def _get_binary_name() -> str:
    return os.path.basename(inspect.stack()[-1][1])


class OSSysLogHandler(logging.Handler):
    """Syslog based handler. Only available on UNIX-like platforms."""

    def __init__(self, facility: int | None = None) -> None:
        # Default values always get evaluated, for which reason we avoid
        # using 'syslog' directly, which may not be available.
        facility = facility if facility is not None else syslog.LOG_USER
        if not syslog:
            raise RuntimeError("Syslog not available on this platform")
        super().__init__()
        binary_name = _get_binary_name()
        syslog.openlog(binary_name, 0, facility)

    def emit(self, record: logging.LogRecord) -> None:
        priority = SYSLOG_MAP.get(record.levelname, 7)
        message = self.format(record)
        syslog.syslog(priority, message)


class OSJournalHandler(logging.Handler):
    custom_fields = (
        'project_name',
        'project_id',
        'user_name',
        'user_id',
        'request_id',
    )

    def __init__(self, facility: int | None = None):
        if not journal:
            raise RuntimeError("Systemd bindings do not exist")

        if not facility:
            if not syslog:
                raise RuntimeError("syslog is not available on this platform")
            facility = syslog.LOG_USER

        # Do not use super() unless type(logging.Handler) is 'type'
        # (i.e. >= Python 2.7).
        logging.Handler.__init__(self)
        self.binary_name = _get_binary_name()
        self.facility = facility

    def emit(self, record: logging.LogRecord) -> None:
        priority = SYSLOG_MAP.get(record.levelname, 7)
        message = self.format(record)

        extras = {
            'CODE_FILE': record.pathname,
            'CODE_LINE': record.lineno,
            'CODE_FUNC': record.funcName,
            'THREAD_NAME': record.threadName,
            'PROCESS_NAME': record.processName,
            'LOGGER_NAME': record.name,
            'LOGGER_LEVEL': record.levelname,
            'SYSLOG_IDENTIFIER': self.binary_name,
            'PRIORITY': priority,
            'SYSLOG_FACILITY': self.facility,
        }

        if record.exc_info:
            # Cache the traceback text to avoid converting it multiple times
            # (it's constant anyway)
            if not record.exc_text and self.formatter is not None:
                record.exc_text = self.formatter.formatException(
                    record.exc_info
                )
        if record.exc_text:
            extras['EXCEPTION_INFO'] = record.exc_text
            # Leave EXCEPTION_TEXT for backward compatibility
            extras['EXCEPTION_TEXT'] = record.exc_text

        for field in self.custom_fields:
            value = record.__dict__.get(field)
            if value:
                extras[field.upper()] = value

        journal.send(message, **extras)


class ColorHandler(_StreamHandler):
    """Log handler that sets the 'color' key based on the level

    To use, include a '%(color)s' entry in the logging_context_format_string.
    There is also a '%(reset_color)s' key that can be used to manually reset
    the color within a log line.
    """

    LEVEL_COLORS: dict[int, str] = {
        _TRACE: '\033[00;35m',  # MAGENTA
        logging.DEBUG: '\033[00;32m',  # GREEN
        logging.INFO: '\033[00;36m',  # CYAN
        _AUDIT: '\033[01;36m',  # BOLD CYAN
        logging.WARNING: '\033[01;33m',  # BOLD YELLOW
        logging.ERROR: '\033[01;31m',  # BOLD RED
        logging.CRITICAL: '\033[01;31m',  # BOLD RED
    }

    def format(self, record: logging.LogRecord) -> str:
        record.color = self.LEVEL_COLORS[record.levelno]
        return logging.StreamHandler.format(self, record) + '\033[00m'



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg

from oslo_log import versionutils

_DEFAULT_LOG_DATE_FORMAT = "%Y-%m-%d %H:%M:%S"

DEFAULT_LOG_LEVELS = [
    'amqp=WARN',
    'boto=WARN',
    'sqlalchemy=WARN',
    'suds=INFO',
    'oslo.messaging=INFO',
    'oslo_messaging=INFO',
    'iso8601=WARN',
    'requests.packages.urllib3.connectionpool=WARN',
    'urllib3.connectionpool=WARN',
    'websocket=WARN',
    'requests.packages.urllib3.util.retry=WARN',
    'urllib3.util.retry=WARN',
    'keystonemiddleware=WARN',
    'routes.middleware=WARN',
    'stevedore=WARN',
    'taskflow=WARN',
    'keystoneauth=WARN',
    'oslo.cache=INFO',
    'oslo_policy=INFO',
    'dogpile.core.dogpile=INFO',
]

_IGNORE_MESSAGE = "This option is ignored if log_config_append is set."

common_cli_opts = [
    cfg.BoolOpt(
        'debug',
        short='d',
        default=False,
        mutable=True,
        help='If set to true, the logging level will be set to '
        'DEBUG instead of the default INFO level.',
    ),
]

logging_cli_opts = [
    cfg.StrOpt(
        'log-config-append',
        metavar='PATH',
        deprecated_name='log-config',
        mutable=True,
        help='The name of a logging configuration file. This file '
        'is appended to any existing logging configuration '
        'files. For details about logging configuration files, '
        'see the Python logging module documentation. Note that '
        'when logging configuration files are used then all '
        'logging configuration is set in the configuration file '
        'and other logging configuration options are ignored '
        '(for example, log-date-format).',
    ),
    cfg.StrOpt(
        'log-date-format',
        default=_DEFAULT_LOG_DATE_FORMAT,
        metavar='DATE_FORMAT',
        help='Defines the format string for %%(asctime)s in log '
        'records. Default: %(default)s . ' + _IGNORE_MESSAGE,
    ),
    cfg.StrOpt(
        'log-file',
        metavar='PATH',
        deprecated_name='logfile',
        help='(Optional) Name of log file to send logging output to. '
        'If no default is set, logging will go to stderr as '
        'defined by use_stderr. ' + _IGNORE_MESSAGE,
    ),
    cfg.StrOpt(
        'log-dir',
        deprecated_name='logdir',
        help='(Optional) The base directory used for relative log_file '
        ' paths. ' + _IGNORE_MESSAGE,
    ),
    cfg.BoolOpt(
        'use-syslog',
        default=False,
        help='Use syslog for logging. '
        'Existing syslog format is DEPRECATED '
        'and will be changed later to honor RFC5424. ' + _IGNORE_MESSAGE,
    ),
    cfg.BoolOpt(
        'use-journal',
        default=False,
        help='Enable journald for logging. '
        'If running in a systemd environment you may wish '
        'to enable journal support. Doing so will use the '
        'journal native protocol which includes structured '
        'metadata in addition to log messages.' + _IGNORE_MESSAGE,
    ),
    cfg.StrOpt(
        'syslog-log-facility',
        default='LOG_USER',
        help='Syslog facility to receive log lines. ' + _IGNORE_MESSAGE,
    ),
    cfg.BoolOpt(
        'use-json',
        default=False,
        help='Use JSON formatting for logging. ' + _IGNORE_MESSAGE,
    ),
]

generic_log_opts = [
    cfg.BoolOpt(
        'use_stderr',
        default=False,
        help='Log output to standard error. ' + _IGNORE_MESSAGE,
    ),
    cfg.BoolOpt(
        'log_color',
        default=False,
        help='(Optional) Set the \'color\' key according to log '
        'levels. This option takes effect only when logging '
        'to stderr or stdout is used. ' + _IGNORE_MESSAGE,
    ),
    cfg.IntOpt(
        'log_rotate_interval',
        default=1,
        help='The amount of time before the log files are rotated. '
        'This option is ignored unless log_rotation_type is set '
        'to "interval".',
    ),
    cfg.StrOpt(
        'log_rotate_interval_type',
        choices=['Seconds', 'Minutes', 'Hours', 'Days', 'Weekday', 'Midnight'],
        ignore_case=True,
        default='days',
        help='Rotation interval type. The time of the last file '
        'change (or the time when the service was started) is '
        'used when scheduling the next rotation.',
    ),
    cfg.IntOpt(
        'max_logfile_count',
        default=30,
        help='Maximum number of rotated log files.',
    ),
    cfg.IntOpt(
        'max_logfile_size_mb',
        default=200,
        help='Log file maximum size in MB. This option is ignored if '
        '"log_rotation_type" is not set to "size".',
    ),
    cfg.StrOpt(
        'log_rotation_type',
        default='none',
        choices=[
            ('interval', 'Rotate logs at predefined time intervals.'),
            ('size', 'Rotate logs once they reach a predefined size.'),
            ('none', 'Do not rotate log files.'),
        ],
        ignore_case=True,
        help='Log rotation type.',
    ),
]

log_opts = [
    cfg.StrOpt(
        'logging_context_format_string',
        default='%(asctime)s.%(msecs)03d %(process)d %(levelname)s '
        '%(name)s [%(global_request_id)s %(request_id)s '
        '%(user_identity)s] %(instance)s%(message)s',
        help='Format string to use for log messages with context. '
        'Used by oslo_log.formatters.ContextFormatter',
    ),
    cfg.StrOpt(
        'logging_default_format_string',
        default='%(asctime)s.%(msecs)03d %(process)d %(levelname)s '
        '%(name)s [-] %(instance)s%(message)s',
        help='Format string to use for log messages when context is '
        'undefined. '
        'Used by oslo_log.formatters.ContextFormatter',
    ),
    cfg.StrOpt(
        'logging_debug_format_suffix',
        default='%(funcName)s %(pathname)s:%(lineno)d',
        help='Additional data to append to log message when logging '
        'level for the message is DEBUG. '
        'Used by oslo_log.formatters.ContextFormatter',
    ),
    cfg.StrOpt(
        'logging_exception_prefix',
        default='%(asctime)s.%(msecs)03d %(process)d ERROR %(name)s '
        '%(instance)s',
        help='Prefix each line of exception output with this format. '
        'Used by oslo_log.formatters.ContextFormatter',
    ),
    cfg.StrOpt(
        'logging_user_identity_format',
        default='%(user)s %(project)s %(domain)s %(system_scope)s '
        '%(user_domain)s %(project_domain)s',
        help='Defines the format string for %(user_identity)s that '
        'is used in logging_context_format_string. '
        'Used by oslo_log.formatters.ContextFormatter',
    ),
    cfg.ListOpt(
        'default_log_levels',
        default=DEFAULT_LOG_LEVELS,
        help='List of package logging levels in logger=LEVEL pairs. '
        + _IGNORE_MESSAGE,
    ),
    cfg.BoolOpt(
        'publish_errors',
        default=False,
        help='Enables or disables publication of error events.',
    ),
    # NOTE(mikal): there are two options here because sometimes we are handed
    # a full instance (and could include more information), and other times we
    # are just handed a UUID for the instance.
    cfg.StrOpt(
        'instance_format',
        default='[instance: %(uuid)s] ',
        help='The format for an instance that is passed with the log message.',
    ),
    cfg.StrOpt(
        'instance_uuid_format',
        default='[instance: %(uuid)s] ',
        help='The format for an instance UUID that is passed with the '
        'log message.',
    ),
    cfg.IntOpt(
        'rate_limit_interval',
        default=0,
        help='Interval, number of seconds, of log rate limiting.',
    ),
    cfg.IntOpt(
        'rate_limit_burst',
        default=0,
        help='Maximum number of logged messages per rate_limit_interval.',
    ),
    cfg.StrOpt(
        'rate_limit_except_level',
        default='CRITICAL',
        choices=['CRITICAL', 'ERROR', 'INFO', 'WARNING', 'DEBUG', ''],
        help='Log level name used by rate limiting. Logs with level '
        'greater or equal to rate_limit_except_level are not '
        'filtered. An empty string means that all levels are '
        'filtered.',
    ),
]


def list_opts() -> list[tuple[str | None, list[cfg.Opt]]]:
    """Returns a list of oslo.config options available in the library.

    The returned list includes all oslo.config options which may be registered
    at runtime by the library.

    Each element of the list is a tuple. The first element is the name of the
    group under which the list of elements in the second element will be
    registered. A group name of None corresponds to the [DEFAULT] group in
    config files.

    The purpose of this is to allow tools like the Oslo sample config file
    generator (oslo-config-generator) to discover the options exposed to users
    by this library.

    :returns: a list of (group_name, opts) tuples
    """
    return [
        (
            None,
            (
                common_cli_opts
                + logging_cli_opts
                + generic_log_opts
                + log_opts
                + versionutils.deprecated_opts
            ),
        )
    ]
